/*
    Copyright 2007,2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#define _LARGE_FILES        // large file support is usually automatic now
#define __USE_LARGEFILE64   // but I prefer to declare all of them
#define __USE_FILE_OFFSET64
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS   64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <zlib.h>

#ifdef WIN32
    #include <windows.h>
    char *get_file(void);
    char *put_file(void);
#else
    #define stricmp strcasecmp
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define VER         "0.1.3"

#define PRINTF64(x) (u32)(x >> 32), (u32)(x)    // I64x, llx? blah
#if defined(__MINGW32__) && defined(_LARGE_FILES)
    #define off_t   off64_t
    #define fopen   fopen64
    #define fseek   fseeko64
    #define ftell   ftello64
#endif



#pragma pack(4)
typedef struct {
    u8      sign[16];       // DAA
    u32     size_offset;    // where starts the list of sizes of the zipped chunks
    u32     b100;           // must be 0x100
    u32     data_offset;    // where starts the zipped chunks
    u32     b1;             // must be 1
    u32     b0;             // ever 0
    u32     chunksize;      // size of each output chunk
    u64     isosize;        // total size of the output ISO
    u64     filesize;       // total size of the DAA file
    u8      zero[16];       // it's ever zero
    u32     crc;            // checksum calculated on the first 0x48 bytes
} daa_t;

#pragma pack(1)
typedef struct {
    u8      n1;             // strange way to store numbers...
    u8      n2;
    u8      n3;
} daa_data_t;

#pragma pack()



u8 *find_ext(u8 *fname, u8 *ext);
FILE *daa_next(void);
void myalloc(u8 **data, unsigned wantsize, unsigned *currsize);
void myfr(FILE *fd, void *data, unsigned size);
void myfw(FILE *fd, void *data, unsigned size);
int unzip(z_stream *z, u8 *in, int insz, u8 *out, int outsz);
void b2l_daa(daa_t *daa);
void b2l_16(u16 *num);
void b2l_32(u32 *num);
void b2l_64(u64 *num);
void std_err(void);
void myexit(void);



int     multi   = 0,
        multinum,
        endian;
char    *filename;



int main(int argc, char *argv[]) {
    z_stream    z;
    daa_data_t  *daa_data;
    daa_t   daa;
    FILE    *fdi,
            *fdo;
    u64     tot;
    u32     type,
            i,
            len,
            daas,
            insz,
            outsz;
    u8      ans[8],
            *filei,
            *fileo,
            *in,
            *out,
            *p;

    setbuf(stdout, NULL);

    fputs("\n"
        "DAA2ISO "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    endian = 1;
    if(*(char *)&endian) endian = 0;

#ifdef WIN32
    if(GetWindowLong(GetForegroundWindow(), GWL_WNDPROC)) {
        argv = malloc(sizeof(char *) * 3);
        argv[1] = get_file();
        argv[2] = put_file();
        argc = 3;
    }
#endif

    if(argc < 3) {
        printf("\n"
            "Usage: %s <input.DAA> <output.ISO>\n"
            "\n", argv[0]);
        myexit();
    }

    filei = argv[1];
    fileo = argv[2];

    printf("- open %s\n", filei);
    fdi = fopen(filei, "rb");
    if(!fdi) std_err();

    printf("- create %s\n", fileo);
    fdo = fopen(fileo, "rb");
    if(fdo) {
        fclose(fdo);
        printf("- the output file already exists, do you want to overwrite it (y/N)? ");
        if(!fgets(ans, sizeof(ans), stdin)) myexit();
        if((ans[0] != 'y') && (ans[0] != 'Y')) myexit();
    }
    fdo = fopen(fileo, "wb");
    if(!fdo) std_err();

    z.zalloc = (alloc_func)0;
    z.zfree  = (free_func)0;
    z.opaque = (voidpf)0;
    if(inflateInit2(&z, -15)) {
        printf("\nError: zlib initialization error\n");
        myexit();
    }

    myfr(fdi, &daa, sizeof(daa));
    b2l_daa(&daa);
    if(strncmp(daa.sign, "DAA", 16)) {
        if(!strncmp(daa.sign, "DAA VOL", 16)) {
            printf("\n"
                "Error: you must choose the first DAA file (*.part01.daa, *.part001.daa or\n"
                "       *.daa) because this is a splitted archive\n");
        } else {
            printf("\nError: wrong DAA signature (%.16s)\n", daa.sign);
        }
        myexit();
    }
    if(crc32(0, (void *)&daa, sizeof(daa) - 4) != daa.crc) {
        printf("\nAlert: the CRC of the DAA header differs than how much it should be\n");
    }
    if((daa.b100 != 0x100) || (daa.b1 != 1)) {
        printf("\nAlert: unknown DAA version (%08x %08x %08x)\n", daa.b100, daa.b1, daa.b0);
    }
    while(ftell(fdi) < daa.size_offset) {
        myfr(fdi, &type, 4);
        b2l_32(&type);
        myfr(fdi, &len,  4);
        b2l_32(&len);
        if(type == 1) {
            multi = 1;  // multi volume, not needed but I want to be sure to catch it in any case
        } else if(type == 3) {
            printf("\nAlert: the input DAA file seems protected by password (not yet supported)\n");
        }
        if(fseek(fdi, len - 8, SEEK_CUR)) std_err();
    }

    fseek(fdi, 0, SEEK_END);
    tot = ftell(fdi);
    if(multi || (tot != daa.filesize)) {
        printf("- multi volume file\n");
        if((p = find_ext(filei, "001.daa"))) {
            multi    = 1;
            multinum = 2;   // the number of the next archive
        } else if((p = find_ext(filei, "01.daa"))) {
            multi    = 2;
            multinum = 2;   // if the current is 1 the next is 2
        } else {
            multi    = 3;
            multinum = 0;
            p = strrchr(filei, '.');
            if(!p) p = filei + strlen(filei);
        }

        len = p - filei;
        filename = malloc(len + 16);
        memcpy(filename, filei, len);
        filename[len] = 0;
    }

    in   = out   = NULL;
    insz = outsz = 0;
    tot  = 0;

    myalloc(&out, daa.chunksize, &outsz);

    daas = daa.data_offset - daa.size_offset;
    daa_data = malloc(daas);
    if(!daa_data) std_err();

    if(fseek(fdi, daa.size_offset, SEEK_SET)) std_err();
    myfr(fdi, daa_data, daas);

    printf("- start unpacking:\n");
    daas /= 3;
    for(i = 0; i < daas; i++) {
        printf("  %03d%%\r", (i * 100) / daas);

        len = (daa_data[i].n1 << 16) | (daa_data[i].n2) | (daa_data[i].n3 << 8);

        myalloc(&in, len, &insz);
        myfr(fdi, in, len);

        len = unzip(&z, in, len, out, outsz);
        myfw(fdo, out, len);
        tot += len;
    }

    printf("  100%%\n"
        "- 0x%08x%08x bytes written\n",
        PRINTF64(tot));

    if(tot != daa.isosize) {
        printf("\nAlert: the size of the ISO should be 0x%08x%08x bytes!\n", PRINTF64(daa.isosize));
    }

    inflateEnd(&z);
    fclose(fdi);
    fclose(fdo);
    if(in)  free(in);
    if(out) free(out);
    myexit();
    return(0);
}



#ifdef WIN32
char *get_file(void) {
    OPENFILENAME    ofn;
    static char     filename[1024];
    static const char   filter[] =
                    "DAA file\0"    "*.daa\0"
                    "(*.*)\0"       "*.*\0"
                    "\0"            "\0";

    filename[0] = 0;
    memset(&ofn, 0, sizeof(ofn));
    ofn.lStructSize     = sizeof(ofn);
    ofn.lpstrFilter     = filter;
    ofn.nFilterIndex    = 1;
    ofn.lpstrFile       = filename;
    ofn.nMaxFile        = sizeof(filename);
    ofn.lpstrTitle      = "Select the input DAA file (or the first multipart) to convert";
    ofn.Flags           = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_LONGNAMES | OFN_EXPLORER | OFN_HIDEREADONLY;

    if(!GetOpenFileName(&ofn)) exit(1);
    return(filename);
}

char *put_file(void) {
    OPENFILENAME    ofn;
    static char     filename[1024];
    static const char   filter[] =
                    "ISO file\0"    "*.iso\0"
                    "(*.*)\0"       "*.*\0"
                    "\0"            "\0";

    filename[0] = 0;
    memset(&ofn, 0, sizeof(ofn));
    ofn.lStructSize     = sizeof(ofn);
    ofn.lpstrFilter     = filter;
    ofn.nFilterIndex    = 1;
    ofn.lpstrFile       = filename;
    ofn.nMaxFile        = sizeof(filename);
    ofn.lpstrTitle      = "Choose the name of the output ISO file to create";
    ofn.Flags           = OFN_PATHMUSTEXIST | OFN_LONGNAMES | OFN_EXPLORER | OFN_HIDEREADONLY;

    if(!GetSaveFileName(&ofn)) exit(1);
    return(filename);
}
#endif



u8 *find_ext(u8 *fname, u8 *ext) {
    int     len,
            extlen;
    u8      *ret;

    len    = strlen(fname);
    extlen = strlen(ext);
    ret    = fname + len - extlen;
    if((len >= extlen) && !stricmp(ret, ext)) {
        return(ret);
    }
    return(NULL);
}



FILE *daa_next(void) {
    daa_t   daa;
    FILE    *fd;
    static char
            *toadd = NULL,
            *fmt;

    if(!toadd) {
        toadd = filename + strlen(filename);
        switch(multi) {
            case 1:  fmt = "%03d.daa";      break;
            case 2:  fmt = "%02d.daa";      break;
            default: fmt = ".d%02d";        break;
        }
    }

    sprintf(toadd, fmt, multinum);
    printf("  open %s\n", filename);
    fd = fopen(filename, "rb");
    if(!fd) std_err();

    myfr(fd, &daa, sizeof(daa));
    b2l_daa(&daa);
    if(strncmp(daa.sign, "DAA VOL", 16)) {
        printf("\nError: wrong DAA VOL signature (%.16s)\n", daa.sign);
        myexit();
    }
    if(fseek(fd, daa.size_offset, SEEK_SET)) std_err();

    multinum++;
    return(fd);
}



void myalloc(u8 **data, unsigned wantsize, unsigned *currsize) {
    if(wantsize <= *currsize) return;
    *data = realloc(*data, wantsize);
    if(!*data) std_err();
    *currsize = wantsize;
}



void myfr(FILE *fd, void *data, unsigned size) {
    int     len;

    len = fread(data, 1, size, fd);
    if(len == size) return;
    if(!multi) {
        printf("\nError: incomplete input file, can't read %u bytes\n", size);
        myexit();
    }
    fclose(fd);
    fd = daa_next();
    myfr(fd, data + len, size - len);
}



void myfw(FILE *fd, void *data, unsigned size) {
    if(fwrite(data, 1, size, fd) == size) return;
    printf("\nError: problems during the writing of the output file\n");
    myexit();
}



int unzip(z_stream *z, u8 *in, int insz, u8 *out, int outsz) {
    inflateReset(z);

    z->next_in   = in;
    z->avail_in  = insz;
    z->next_out  = out;
    z->avail_out = outsz;
    if(inflate(z, Z_SYNC_FLUSH) != Z_STREAM_END) {
        printf("\nError: the compressed input is wrong or incomplete\n");
        myexit();
    }
    return(z->total_out);
}



void b2l_daa(daa_t *daa) {
    if(!endian) return;
    b2l_32(&daa->size_offset);
    b2l_32(&daa->b100);
    b2l_32(&daa->data_offset);
    b2l_32(&daa->b1);
    b2l_32(&daa->b0);
    b2l_32(&daa->chunksize);
    b2l_64(&daa->isosize);
    b2l_64(&daa->filesize);
    b2l_32(&daa->crc);
}



void b2l_16(u16 *num) {
    u16     tmp;

    if(!endian) return;

    tmp = *num;
    *num = ((tmp & 0xff00) >> 8) |
           ((tmp & 0x00ff) << 8);
}



void b2l_32(u32 *num) {
    u32     tmp;

    if(!endian) return;

    tmp = *num;
    *num = ((tmp & 0xff000000) >> 24) |
           ((tmp & 0x00ff0000) >>  8) |
           ((tmp & 0x0000ff00) <<  8) |
           ((tmp & 0x000000ff) << 24);
}



void b2l_64(u64 *num) {
    u64     tmp;

    if(!endian) return;

    tmp = *num;
    *num = ((u64)(tmp & (u64)0xff00000000000000ULL) >> 56) |
           ((u64)(tmp & (u64)0x00ff000000000000ULL) >> 40) |
           ((u64)(tmp & (u64)0x0000ff0000000000ULL) >> 24) |
           ((u64)(tmp & (u64)0x000000ff00000000ULL) >>  8) |
           ((u64)(tmp & (u64)0x00000000ff000000ULL) <<  8) |
           ((u64)(tmp & (u64)0x0000000000ff0000ULL) << 24) |
           ((u64)(tmp & (u64)0x000000000000ff00ULL) << 40) |
           ((u64)(tmp & (u64)0x00000000000000ffULL) << 56);
}



void std_err(void) {
    perror("\nError");
    myexit();
}



void myexit(void) {
#ifdef WIN32
    if(GetWindowLong(GetForegroundWindow(), GWL_WNDPROC)) {
        printf("\nPress RETURN to quit");
        while(fgetc(stdin) != '\n');
    }
#endif
    exit(1);
}


